#include <Cg/cg.h>
#include <Cg/cgGL.h>

CGcontext cgContext=NULL;
//CGprogram vProg;
//CGprogram fProg;
CGprofile vProfile;
CGprofile fProfile;

/** linked list with loaded shaders */
typedef struct _shaderlist {
 CGprogram cp;
 char type;//v/f
 char *file;
 char *name;
 struct _shaderlist *next;
} shaderlist;

/** list of shaders in application */
shaderlist *shlist;

/** get program from shader list, loading it if not in memory
   @param type type of shader ('f'=fragment, 'v'=vertex)
   @param sfile name of file with shader
   @param sname name of function within the file to be loaded
 */
CGprogram getprogram(char type,const char *sfile,const char *sname) {
 shaderlist *t=shlist;
 while (t) {
  if (t->type==type && strcasecmp(t->file,sfile)==0 && strcasecmp(t->name,sname)==0) return t->cp;
  t=t->next;
 }
 t=(shaderlist *)malloc0(sizeof(shaderlist));
 t->type=type;
 t->next=shlist;
 t->file=strdup(sfile);
 t->name=strdup(sname);
 t->cp=cgCreateProgramFromFile(cgContext,CG_SOURCE,sfile,type=='f'?fProfile:vProfile,sname,0);
 if (t->cp==NULL) fatalf("Failed loading program %s from file %s (%s)\n",sname,sfile,cgGetErrorString(cgGetError()));
 cgGLLoadProgram(t->cp);
 printf("Loading shader: %s (%s)\n",sfile,sname);
 glerr("getprogram");
 shlist=t;
 return shlist->cp;
}

/** Initialize Cg environment. */
void createCg () {
  cgContext=cgCreateContext();
//  cgGLSetManageTextureParameters(cgContext,CG_TRUE);
  if (cgContext==NULL) fatalf("Failed to create Cg context: %s",cgGetErrorString(cgGetError()));
  vProfile=cgGLGetLatestProfile(CG_GL_VERTEX);
  if (vProfile==CG_PROFILE_UNKNOWN)  fatalf("Invalid Cg vertex profile: %s",cgGetErrorString(cgGetError()));
  cgGLSetOptimalOptions(vProfile);          // optimizing options for compiler
//  vProg=getprogram('v',shaderfile,"vsPassThrough");
  fProfile=cgGLGetLatestProfile(CG_GL_FRAGMENT);
  if (fProfile==CG_PROFILE_UNKNOWN) fatalf("Invalid Cg fragment profile: %s",cgGetErrorString(cgGetError()));
  cgGLSetOptimalOptions(fProfile);          // optimizing options for compiler
//  fProg=getprogram('f',shaderfile,"fsPhong");
}

/** Quit Cg environment. */
void destroyCg() {
//  cgDestroyProgram(vProg);
//  cgDestroyProgram(fProg);
  cgDestroyContext(cgContext);
  cgContext=NULL;
}

#ifndef WIN32
#define wglGetProcAddress(x) glXGetProcAddressARB((const GLubyte *)(x))
#endif
