#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef WIN32
#include <windows.h>
#else
#include <GL/glx.h>
//#include <GL/glxext.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <math.h>
#include <assert.h>
#include "glext.h"
#include "util.h"
#include "texture.h"
#include "cgsh.h"

#define apptitle "Flame"

#ifdef _MSC_VER
#define strncasecmp stricmp
#endif

#define VERSION "0.1.0"

#define GLUT_MOUSE 10000	//mouse button virtual key constant
int mousex,mousey,omousex=-1,omousey=-1;	//mouse position
int maxx,maxy;		//X/Y dimensions of window
int ms_last;	//last/current miliseconds needed for frame
float ms_curr;

int wireframe,help;

double angle1,angle2=0.5;

typedef float float3[3];
typedef unsigned int uint;

float3 eye;	//eye coordinates
float3 eye_d;	//direction of view "front" vector
float3 eye_u;	//direction of view "up" vector

float3 fogcolor={0.25,0.25,0.8};

float sensitivity=0.015;	//mouse sensitivity
float kb_sensitivity=20;	//keyboard sensitivity
float fov=90;			//Field of view
float radius=40;		//camera radius
float clipfront=0.001;		//near clipping plane distance
float clipback=250;		//far clipping plane distance
/** name of texture containing font */
char *font_texture_name=strdup("font.tga");
/** name of file containing scene definition */
char *scene_filename=strdup("scene.txt");
/** if true, framebuffer extension will be used */
int use_fbo=1;

uint buffer2;

/** return number of miliseconds since start */
uint gettickcount(void) {
 return glutGet(GLUT_ELAPSED_TIME);
}

#define DIM3D 32

typedef int tex3d[DIM3D][DIM3D][DIM3D];//int32 3d texture

tex3d t3data[2];

GLuint t3[2];

/** generate 3d texture with noise
 @param tex3 pointer to 3dtexture of dimensions DIM3D ^3
 */
void generate3d(tex3d *tex3) {
 for (int x=0;x<DIM3D;x++){
  for (int y=0;y<DIM3D;y++){
   for (int z=0;z<DIM3D;z++){
    int c=rand()%256;
    c=c|c<<8|c<<16|c<<24;
    (*tex3)[x][y][z]=c;
   }
  }
 }
}

typedef unsigned char ub4[4];	//4 unsigned bytes
typedef float f3[3];		//3 floats
typedef float f2[2];		//2 floats

ub4 white={255,255,255,255}; //white color
ub4 black={0,0,0,255}; //white color

/** texture pointing to the second framebuffer (other one than rendering target) */
GLuint fbo_tex;

/** parameter list for shaders */
typedef struct _paramlist {
 char *name;
 CGparameter cp;
 char *varname;
 int vcount;
 float value[4];
 GLuint texvalue;
 struct _paramlist *next;
} paramlist;

/** class containing shader (CGProgram) with list of parameters.
 One CGProgram can be contained in different tshaders with different parameters.
 Class can contain either vertex or fragment shader
 If name of file and function is empty, the shader is just used as a parameter list
 */
class tshader {
 /** true if no shader inside (just a parameter list) */
 int dummy;
 /** parameter list for shader */
 paramlist *param;
 /** type of shader ('f'=fragment, 'v'=vertex ) */
 char styp;
 /** common functions used when adding parameter to shader */
 void addparam_common(paramlist *p,const char *name,int mayfail=0) {
  p->cp=NULL;
  if (!dummy)  {
   p->cp=cgGetNamedParameter(program,name);
   if (p->cp==NULL) {
    if (mayfail) {
     if (p->varname) free(p->varname);
     if (p==param) { //that was something new - put it away
      param=p->next;
      free(p);
     }
     return;//nothing added
    } else {
     fatalf("Parameter %s does not exist in program\n",name);
    }
   }
  }
  p->name=strdup(name);
 }
 /** the shader itself */
 CGprogram program;
public:

 /** return shader type ('v' = vertex, 'f' = fragment) */
 char type() {
  return styp;
 }
 /** bind contained program to openGL */
 void bindprogram() {
  assert(program!=NULL);
  glerr("before tshader.bindprogram");
  cgGLBindProgram(program);
  glerr("after tshader.bindprogram");
 }
 /** construct shader from file 
  @param typ  type of shader ('f'=frragment, 'v'=vertex )
  @param file name of file with shader
  @param name name of function within the file to be loaded
  */
 tshader(char typ, const char *file, const char *name) {
  param=NULL;
  dummy=(*file==0 && *name==0);
  if (!dummy) program=getprogram(typ,file,name);
  styp=typ;
 }

 /** get parameter with given name */
 paramlist *getparam(const char *name) {
  paramlist *p=param;
  while (p){
   if (strcmp(p->name,name)==0) {
    return p;
   }
   p=p->next;
  }  
  return NULL;
 }

 /** DEBUG: print parameter list to STDOUT */
 void printparam(void) {
  paramlist *p=param;
  printf("Parameter list\n");
  while (p){
   printf("PARAM %s $%s (%d)\n",p->name,p->varname,p->vcount);
   p=p->next;
  }  
 }

 /** This function will either get parameter from the list if it exist and clear it or allocate a new parameter */
 paramlist *get_empty_param(const char *name) {
  paramlist *p=getparam(name);
  if (!p) {
   p=(paramlist *)malloc0(sizeof(paramlist));
   p->next=param;
   param=p;
  } else {
   paramlist *bk=p->next;
   if (p->name) free(p->name);
   if (p->varname) free(p->varname);
   memset(p,0,sizeof(paramlist));
   p->next=bk;
  }
  return p;
 }

 /** add parameter to shader
  If the parameter exist, it is overwritten
  @param name name of parameter
  @param fcount number of float parameters in array (1..4)
  @param value array with parameters (float)
  @param mayfail if true and parameter does not exist, function will silently fail
 */
 void addparam(const char *name,int fcount,float *value,int mayfail=0) {
  paramlist *p=get_empty_param(name);
  assert(fcount>=1);
  assert(fcount<=4);
  p->vcount=fcount;
  memcpy(p->value,value,sizeof(float)*fcount);
  addparam_common(p,name,mayfail);
 }
 /** add dynamic parameter to shader
  If the parameter exist, it is overwritten
  @param name name of parameter
  @param var Name of variable. This parameter will be re-evaluated each frame.
  @param mayfail if true and parameter does not exist, function will silently fail
 */
 void addparam(const char *name,const char *var,int mayfail=0) {
  paramlist *p=get_empty_param(name);
  p->varname=strdup(var);
  addparam_common(p,name,mayfail);
 }

 /** Copy all parameters from this shader to another shader, if they are present in target shader */
 void putparams(tshader *t) {
  paramlist *p=param;
  while (p){
   if (p->varname) {
    t->addparam(p->name,p->varname,1);
   } else {
    t->addparam(p->name,p->vcount,p->value,1);
   }
   p=p->next;
  }
 }

 /** called each frame. Will free parameters for given shader */
 void unsetparams() {
  paramlist *p=param;
  while (p){
   if (p->vcount==-1) { //texture
    cgGLDisableTextureParameter(p->cp);
   }
   p=p->next;
  }
 }

 /** called each frame. Will set up parameters for given shader */
 void setparams() {
  paramlist *p=param;
  while (p){
   if (p->varname) {
    if (strcasecmp(p->varname,"modelview_projection")==0) { //VARIABLE
     p->vcount=16;
     cgGLSetStateMatrixParameter(p->cp,CG_GL_MODELVIEW_PROJECTION_MATRIX,CG_GL_MATRIX_IDENTITY);
    } else if (strcasecmp(p->varname,"modelview")==0) { //VARIABLE
     p->vcount=16;
     cgGLSetStateMatrixParameter(p->cp,CG_GL_MODELVIEW_MATRIX,CG_GL_MATRIX_IDENTITY);
    } else if (strcasecmp(p->varname,"texture")==0) { //VARIABLE
     p->vcount=16;
     cgGLSetStateMatrixParameter(p->cp,CG_GL_TEXTURE_MATRIX,CG_GL_MATRIX_IDENTITY);
    } else if (strcasecmp(p->varname,"projection")==0) { //VARIABLE
     p->vcount=16;
     cgGLSetStateMatrixParameter(p->cp,CG_GL_PROJECTION_MATRIX,CG_GL_MATRIX_IDENTITY);
    } else if (strcasecmp(p->varname,"eye")==0) { //VARIABLE
     for (int i=0;i<3;i++) p->value[i]=eye[i];
     p->value[3]=1;
     p->vcount=4;
    } else if (strcasecmp(p->varname,"eye_d")==0) { //VARIABLE
     for (int i=0;i<3;i++) p->value[i]=eye_d[i];
     p->value[3]=1;
     p->vcount=4;
    } else if (strcasecmp(p->varname,"eye_u")==0) { //VARIABLE
     for (int i=0;i<3;i++) p->value[i]=eye_u[i];
     p->value[3]=1;
     p->vcount=4;
    } else if (strcasecmp(p->varname,"time")==0) { //VARIABLE
     p->value[0]=(float)gettickcount();
     p->vcount=1;
    } else if (strcasecmp(p->varname,"framebuffer")==0) { //VARIABLE
     if (!use_fbo) fatalf("Variable $framebuffer is available only if EXT_frambuffer_object extension is available");
     p->texvalue=fbo_tex;
     p->vcount=-1;
    } else if (strcasecmp(p->varname,"tex3d_1")==0) { //VARIABLE
     p->texvalue=t3[0];
     p->vcount=-1;
    } else if (strcasecmp(p->varname,"tex3d_2")==0) { //VARIABLE
     p->texvalue=t3[1];
     p->vcount=-1;
    } else fatalf("Unknown dynamic variable %s\n",p->varname);
   }
   switch (p->vcount) {
    case -1://texture
            cgGLSetTextureParameter(p->cp,p->texvalue);
            cgGLEnableTextureParameter(p->cp);
            break;
    case 16: break;//that was matrix
    case 4: cgGLSetParameter4fv(p->cp,p->value);break;
    case 3: cgGLSetParameter3fv(p->cp,p->value);break;
    case 1: cgGLSetParameter1fv(p->cp,p->value);break;
    case 2: cgGLSetParameter2fv(p->cp,p->value);break;
    default: {
     fatalf("vcount=%d,param=%s",p->vcount,p->name);
    }
   }
   p=p->next;
  }
 }
};

/** normalize vector of 3 floats
 @param vec vector to normalize
 */
void normalize3 (float *vec) {
 float size=sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]);
 if (size==0) return;
 vec[0]/=size;
 vec[1]/=size;
 vec[2]/=size;
}


/** pointer to shader class */
typedef tshader *shader;

/** class containing model in B-representation as vertex, triangle and texcoord lists */
class model {
protected:
 int vtxnum;//pocet vertexu (vtx)
 f3 *vtx;//vertexy
 int trinum;//pocet trojuhelniku (tri)
 int *tri;//trojuhelniky -> index do vertexu
 int texnum;//pocet texturovacich souradnic (tex)
 f2 *tex;//texture coordinates
 f3 *tex3;//texture coordinates (3d)
 int *texi;//index do texture coordinates (pro kazdy bod v trojuhelniku)
 int normals;//pokud je 1, pouziji se normaly
 f3 *nor;//normaly (stejny pocet jako vertexy)
 /** call after setting vtxnum,trinum and texnum. Will allocate memory for vertex,triangles and texcoord data */
 void alloc() {
  if (vtxnum) {
   vtx=(f3*)malloc(sizeof(f3)*vtxnum);
  } else vtx=NULL;
  if (trinum) {
   tri=(int*)malloc(sizeof(int)*trinum*3);
   texi=(int*)malloc(sizeof(int)*trinum*3);
  } else { tri=NULL;texi=NULL; }
  if (texnum) {
   tex=(f2*)malloc(sizeof(f2)*texnum);
   tex3=(f3*)malloc(sizeof(f3)*texnum);
  } else { tex=NULL;tex3=NULL; }
  if (normals) {
   nor=(f3*)malloc(sizeof(f3)*vtxnum);
  } else nor=NULL;
 }
 /** construct model with given number of vertexes, triangles and texture coordinates */
 model(int vertexes,int tris,int texi,int norms=0){
  vtxnum=vertexes;
  trinum=tris;
  texnum=vtxnum;
  normals=norms;
  alloc(); 
 }
public:
 /** draw this model unsing OpenGL commands
   @param use3d Use 3d coordinates instead of 2d
 */
 virtual void draw(int use3d) {
  if (wireframe) {	//wireframe
   glDisable(GL_TEXTURE_2D);
//   glColor4b(255,0,0,255);
   glBegin(GL_LINES);
   for(int i=0;i<trinum;i++) {
    glVertex3fv(vtx[tri[i*3]]);
    glVertex3fv(vtx[tri[i*3+1]]);
    glVertex3fv(vtx[tri[i*3+1]]);
    glVertex3fv(vtx[tri[i*3+2]]);
    glVertex3fv(vtx[tri[i*3+2]]);
    glVertex3fv(vtx[tri[i*3]]);
   }
   glEnd();  
  } else {		//normal
   glBegin(GL_TRIANGLES);
   for(int i=0;i<trinum*3;i++) {
    if (use3d) {
     glTexCoord3fv(tex3[texi[i]]);
    } else {
     glTexCoord2fv(tex[texi[i]]);
    }
    if (normals) {
     glNormal3fv(nor[tri[i]]);
    }
    glVertex3fv(vtx[tri[i]]);
   }
   glEnd();
  }
 }

 /** draw this model unsing OpenGL commands
 */
 virtual void draw() {
  draw(0);
 }
 /** destructor - free memory used by model */
 virtual ~model() {
  if (vtx) free(vtx);
  if (tex) {free(tex);free(tex3);}
  if (tri) free(tri);
  if (nor) free(nor);
 }
};

/** reset material - remove textures from texturing units in shaders */
void material_reset(texture *tex,shader vshader,shader fshader) {
 if (vshader) vshader->unsetparams();
 if (fshader) fshader->unsetparams();
 if (tex && tex->is3d) {
  glDisable(GL_TEXTURE_3D);
 }
}

/** set material for OpenGL - color, texture, pixel and framgent shaders */
void material(texture *tex,shader vshader,shader fshader,ub4 color) {
 if (tex) {
  if (tex->is3d) {
   glEnable(GL_TEXTURE_3D);
   glDisable(GL_TEXTURE_2D);
//   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_3D,tex->id);
  } else {
   glEnable(GL_TEXTURE_2D);
   glBindTexture(GL_TEXTURE_2D,tex->id);
  }
 } else {
  glDisable(GL_TEXTURE_2D);
 }
 if (wireframe) {
  glColor4ub(color[0],color[1],color[2],255);
 } else glColor4ubv(color);
 if (vshader) {
  cgGLEnableProfile(vProfile);
  vshader->bindprogram();
  vshader->setparams();
 } else {
  cgGLDisableProfile(vProfile);
 }
 if (fshader) {
  cgGLEnableProfile(fProfile);
  fshader->bindprogram();
  fshader->setparams();
 } else {
  cgGLDisableProfile(fProfile);
 }
}

/** return size of the vector squared */
float size2(float3 v) {
 return v[0]*v[0]+v[1]*v[1]+v[2]*v[2];
}

/** model extended with color, texture, pixel and framgent shaders information */
class item: public model{
protected:
 shader vsh,fsh;
 texture *t; 
 ub4 color;
 f3 translation;
 f3 distance;
 f3 rot;
 float angle;
public:
 /** cached distance set by recalc_distance2() */
 float dist2;
 /** if true, items is transparent and order of drawing matters */
 int transparent;
 /** if true (default), back faces will be culled for this model */
 int cull;
 /** if true (default), z-buffer will be updated. If false,
  z-buffer values will not be updated while drawing this model. */
 int zwrite;
 /** if true, z-buffer will be tested (default). If false,
  z-buffer will be ignored. */
 int d_test;
 /** pass-thru constructor (see model) */
 item(int vertexes,int tris,int texi,int normals=0) : model(vertexes,tris,texi,normals) {
  for (int i=0;i<4;i++) color[i]=255;
  move(0,0,0);
  rotate(0,0,0,0);
  cull=1;
  zwrite=1;
  d_test=1;
  transparent=0;
  vsh=NULL;fsh=NULL;t=NULL;
 }
 /** set translation of object to (x,y,z) */
 void move(float x,float y,float z) {
  translation[0]=x;
  translation[1]=y;
  translation[2]=z;
  for (int i=0;i<3;i++) distance[i]=translation[i];
 }
 /** set "translation" used to calculate distance of object to (x,y,z) */
 void setdistance(float x,float y,float z) {
  distance[0]=x;
  distance[1]=y;
  distance[2]=z;
 }
 /** set rotation of object to angle around axis (x,y,z). Rotation is applied after translation */
 void rotate(float _angle,float x,float y,float z) {
  rot[0]=x;
  rot[1]=y;
  rot[2]=z;
  angle=_angle;
 }
 /** draw this model unsing OpenGL commands */
 virtual void draw() {
  if (cull) glEnable(GL_CULL_FACE);
  else glDisable(GL_CULL_FACE);
  if (!zwrite) glDepthMask(0);
  if (!d_test) glDisable(GL_DEPTH_TEST);
  glPushMatrix();
  glTranslatef(translation[0],translation[1],translation[2]);
  glRotatef(angle,rot[0],rot[1],rot[2]);
  material(t,vsh,fsh,color);
  model::draw((t && t->is3d));
  material_reset(t,vsh,fsh);
  glPopMatrix();
  if (!d_test) glEnable(GL_DEPTH_TEST);
  if (!zwrite) glDepthMask(1);
 }
 /** set color for this model
  @param _color color to set (in (R,G,B,A) format)
  */
 void setcolor(ub4 _color) {
  for (int i=0;i<4;i++) color[i]=_color[i];
 }
 /** set color for this model
  @param _color color to set (in 0xRRGGBBAA format)
 */
 void setcolor(int _color) {
  for (int i=0;i<4;i++) color[i]=(_color <<i*8) & 0xff;
 }
 /** set texture used for this model
  @param tex texture to set
  */
 void settexture(texture *tex) {
  t=tex;
 }
 /** set fragment/vertex shader for this model
   @param type type of shader ('f'=fragment, 'v'=vertex)
   @param sh shader with parameters
  */
 void setshader(char type,shader sh) {
  if (type=='f') fsh=sh; else vsh=sh;
 }
 /** return squared distance of object (between "translation" and "eye") based on eye coordinates and direction of view */
 float distance2(float3 eyep,float3 d) const {
  int i;
  float3 eyex;
  for (i=0;i<3;i++) eyex[i]=distance[i]-eyep[i];
  //eye=eye-relative distance to object
  return size2(eyex);
 }
/* ~item() {
 }*/
 /** recalculate squared distance from eye and put it in member variable dist2 */
 void recalc_distance2() {
  dist2=distance2(eye,eye_d);
 }
};

/** empty model (no vertex/surface information) */
class empty: public item {
public:
 empty(): item(0,0,0) {
 }
};

/** flat - quad paralled with x and y axis*/
class flat: public item {
public:
 /** constructor for axis-aligned quad from (x1,y1,z) to (x2,y2,z) */
 flat(float x1,float y1,float x2,float y2,float z) : item(4,2,4) {
  vtx[0][0]=x1;  vtx[0][1]=y1;  vtx[0][2]=z;
  vtx[1][0]=x2;  vtx[1][1]=y1;  vtx[1][2]=z;
  vtx[2][0]=x2;  vtx[2][1]=y2;  vtx[2][2]=z;
  vtx[3][0]=x1;  vtx[3][1]=y2;  vtx[3][2]=z;
  tex[0][0]=0;  tex[0][1]=0;
  tex[1][0]=1;  tex[1][1]=0;
  tex[2][0]=1;  tex[2][1]=1;
  tex[3][0]=0;  tex[3][1]=1;
  texi[0]=tri[0]=0;
  texi[1]=tri[1]=1;
  texi[2]=tri[2]=2;
  texi[3]=tri[3]=2;
  texi[4]=tri[4]=3;
  texi[5]=tri[5]=0;
 }
};

/** special model - a house */
class house:public empty {
 GLUquadricObj* q;
public:
 /** default and only constructor */
 house() : empty() {
  q=gluNewQuadric();
  gluQuadricTexture(q,GL_TRUE);
 }
 /** draw the house using OpenGL */
 void draw() {
  empty::draw();//textures, etc
  glPushMatrix();
  if (wireframe) gluQuadricDrawStyle(q,GLU_LINE);
  else gluQuadricDrawStyle(q,GLU_FILL);
  glRotatef(0,0,1,pi/4.0);
  gluCylinder(q,20,20,10,4,1);
  glTranslatef(0,0,10);
  gluCylinder(q,20,0,10,4,1);
  glPopMatrix();
 }
};

/** A cylinder caps */
class cylindercaps:public item {
public:
 /** contruct new cylinder caps
  @param xsegments number of segments around the cylinder (circular segments)
  @param width width of cylinder
  @param height height of cylinder
  */ 
 cylindercaps(int xsegments,float width,float height) : item (xsegments*2,(xsegments-2)*2,xsegments*2) {
  int x,z,i=0,t=0;
  float a,at,zt;
  for (z=0;z<2;z++) {
   if (!z) zt=-1; else zt=1;
   for (x=0;x<xsegments;x++) {
    a=(float)x/(float)xsegments*2.0*pi;
    at=1-(float)x/(float)xsegments;
    assert(i<vtxnum);
    assert(i<texnum);
    vtx[i][0]=-width*sin(a);
    vtx[i][1]=-width*cos(a);
    vtx[i][2]=height*(float)z;
    tex[i][0]=sin(a)*zt/2.0+0.5;
    tex[i][1]=cos(a)/2.0+0.5;
    i++;
   }
  }
  for (x=1;x<xsegments-1;x++) {    
   texi[t]=tri[t]=  0;
   t++; 
   texi[t]=tri[t]=  x;
   t++; 
   texi[t]=tri[t]=x+1;
   t++;
  }
  int x1=xsegments;
  for (x=1;x<xsegments-1;x++) {    
   texi[t]=tri[t]=  0+x1;
   t++; 
   texi[t]=tri[t]=x+1+x1;
   t++;
   texi[t]=tri[t]=  x+x1;
   t++; 
  }
 }
};

/** A cylinder (without caps) */
class cylinder:public item {
public:
 /** contruct new cylinder
  @param xsegments number of segments around the cylinder (circular segments)
  @param zsegments number of segments in z-axis (straingt segments)
  @param width width of cylinder
  @param height height of cylinder
  */ 
 cylinder(int xsegments,int zsegments,float width,float height) : item ((xsegments+1)*(zsegments+1),xsegments*zsegments*2,(xsegments+1)*(zsegments+1),1) {
  int x,z,i=0,t=0;
  float a,at,zt;//,at2,zt2;
  for (x=0;x<=xsegments;x++) {
   a=(float)x/(float)xsegments*2.0*pi;
   at=1-(float)x/(float)xsegments;
   for (z=0;z<=zsegments;z++) {
    zt=(float)z/(float)zsegments;
    assert(i<vtxnum);
    assert(i<texnum);
    vtx[i][0]=-width*sin(a);
    vtx[i][1]=-width*cos(a);
    vtx[i][2]=height*(float)z/(float)zsegments;
    nor[i][0]=-cos(a);
    nor[i][1]=+sin(a);
    nor[i][2]=0;
    tex[i][0]=at;
    tex[i][1]=zt;
    tex3[i][0]=sin(a);
    tex3[i][1]=cos(a);
    tex3[i][2]=(float)z/(float)zsegments;
    i++;
   }
  }
  int x1;
  for (x=0;x<xsegments;x++) {
//   at2=(float)(x+1)/(float)xsegments;
//   zt2=(float)(z+1)/(float)zsegments;
   for (z=0;z<zsegments;z++) {
    x1=x+1;
    if (x1==xsegments) x1=0;
    texi[t]=tri[t]=x *(zsegments+1)+z;
    t++; 
    tri[t]=x1*(zsegments+1)+z+1;
    texi[t]=(x+1)*(zsegments+1)+z+1;
    t++; 
    tri[t]=x1*(zsegments+1)+z;
    texi[t]=(x+1)*(zsegments+1)+z;
    t++;
    texi[t]=tri[t]=x *(zsegments+1)+z;
    t++; 
    texi[t]=tri[t]=x *(zsegments+1)+z+1;
    t++; 
    tri[t]=x1*(zsegments+1)+z+1;
    texi[t]=(x+1)*(zsegments+1)+z+1;
    t++; 
   }
  }
 }
};

/** Pseudocylinder - a deformed cylinder.
At exponent=3 it is somewhat rounded rectangle.
*/
class pseudocylinder:public cylinder {
public:
 /** contruct new cylinder
  @param xsegments number of segments around the cylinder (circular segments)
  @param zsegments number of segments in z-axis (straingt segments)
  @param width width of cylinder
  @param height height of cylinder
  @param exponent exponent (exponent=3 -> rounded rectangle, exponent=-INF -> cylinder, should be non-negative integer otherwise errors may occur)
  */ 
 pseudocylinder(int xsegments,int zsegments,float width,float height,float exponent) : cylinder(xsegments, zsegments, width, height) {
  int x,z,i=0;
  float a;
  for (x=0;x<=xsegments;x++) {
   for (z=0;z<=zsegments;z++) {
    a=(float)x/(float)xsegments*2.0*pi;
    vtx[i][0]=-width*(sin(a)+pow(sin(a),exponent));
    vtx[i][1]=-width*(cos(a)+pow(cos(a),exponent));
    nor[i][0]=-cos(a)-pow(cos(a),exponent);
    nor[i][1]=+sin(a)+pow(sin(a),exponent);
    normalize3(nor[i]);
    i++;
   }
  }  
 }
};

/** Cone segment - without caps
*/
class coneseg:public cylinder {
public:
 /** contruct new cone segment
  @param xsegments number of segments around the cylinder (circular segments)
  @param zsegments number of segments in z-axis (straingt segments)
  @param width0 width of cone upside
  @param width1 width of cone downside
  @param height height of cone
  */ 
 coneseg(int xsegments,int zsegments,float width0,float width1,float height) : cylinder(xsegments, zsegments, 1, height) {
  int x,z,i=0;
  float a;
  for (x=0;x<=xsegments;x++) {
   for (z=0;z<=zsegments;z++) {
    a=(float)z/(float)zsegments;
    vtx[i][0]*=(width0*a+width1*(1-a));
    vtx[i][1]*=(width0*a+width1*(1-a));
    i++;
   }
  }  
 }
};

/** center mouse - move mouse pointer in screen center */
void mouseCenter(void) {
 mousex=maxx/2;
 mousey=maxy/2;
 omousex=mousex;
 omousey=mousex;
 glutWarpPointer(mousex,mousey);
}

/** mouse move handler */
void mouseHandler(int x,int y) {
 if (omousex==-1) {
  omousex=mousex=x;
  omousey=mousey=y;
  mouseCenter();
  return;
 }
 omousex=mousex;
 omousey=mousey;
 mousex=x;
 mousey=y;
 //move view
 angle1+=(mousex-omousex)*sensitivity;
 angle2-=(mousey-omousey)*sensitivity;
 if (angle2>1.5) angle2=1.5;
 if (angle2<-1.5) angle2=-1.5;
 //center mouse
 if ((mousex==omousex) && (mousey==omousey)) return;
 mouseCenter();
}

/** GLUT window handle */
int gwin;

/** Quit routine */
void quit(void){
 printf("Quitting...\n");
 glutDestroyWindow(gwin);
 printf("Window destroyed\n");
 destroyCg();
 printf("CG destroyed\n");
 exit(EXIT_SUCCESS);
}

float radiusdelta=0;

/** keyboard press handler*/
void kbd(int c,int is_down) {
 if (is_down) {
  switch (c) {
   case 'q':  quit();  break;
   case 'w':  wireframe=!wireframe;  break;
   case 'h':  help=!help;  break;
   case -GLUT_KEY_UP:  radiusdelta=-1;  break;
   case -GLUT_KEY_DOWN:  radiusdelta=+1;  break;
  }
 } else {
  switch (c) {
   case -GLUT_KEY_UP:  radiusdelta=0;  break;
   case -GLUT_KEY_DOWN:  radiusdelta=0;  break;
  }
 }
}

/** mouse button press/release handler */
void mouseButtonHandler(int button,int is_down,int x,int y) {
 kbd(GLUT_MOUSE+button,is_down==GLUT_DOWN);
}

/** special key press handler */
void skeyDownHandler(int c,int x,int y) {
 kbd(-c,1);
}

/** special key release handler */
void skeyUpHandler(int c,int x,int y) {
 kbd(-c,0);
}

/** key press handler */
void keyDownHandler(unsigned char c, int x,int y) {
 kbd(c,1);
}

/** key release handler */
void keyUpHandler(unsigned char c, int x,int y) {
 kbd(c,0);
}


/** 2 offscreen framebuffers */
GLuint fbuf[2];
/** 2 offscreen renderbuffers */
GLuint rbuf_depth[2];

/** 2 textures holding offscreen framebuffer image */
GLuint fb_image[2];

GLenum fb_buf[2]={GL_COLOR_ATTACHMENT0_EXT,GL_BACK_LEFT};
GLenum screen_buf=GL_BACK_LEFT;

/** Create and return texture with same dimensions as screen */
GLuint createScreenImage(int w,int h) {
 uint res;
 glGenTextures(1,&res);
 glBindTexture(GL_TEXTURE_2D,res);
 glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,w,h,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);
// glCopyTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,0,0,w,h,0);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
 if (!res) fatalf("Cannot create offscreen texture");
 return res;
}

#define FBTYPE GL_TEXTURE_2D
/** create two offscreen buffers */
void initFBO(int w,int h) {
 assert(use_fbo);
 printf("Creating framebuffer objects: %dx%d\n",w,h);
 for(int i=0;i<2;i++) {
  glGenFramebuffersEXT(1,&fbuf[i]);	//framebuffer
  glGenRenderbuffersEXT(1,&rbuf_depth[i]);	//renderbuffer - depth
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbuf[i]);
  fb_image[i]=createScreenImage(w, h);	//texture (framebuffer format) & init the texture

  //attach texture to framebuffer colorbuffer
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,FBTYPE,fb_image[i],0);

  if (i==0) { //only first need a depth buffer
  //initialize depth buffer
   glBindRenderbufferEXT(GL_RENDERBUFFER_EXT,rbuf_depth[i]);
   glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,GL_DEPTH_COMPONENT24,w,h);
   glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,rbuf_depth[i]);
  }

  int status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  switch(status) {
   case GL_FRAMEBUFFER_COMPLETE_EXT:
    break;
   case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
    fatalf("(FBO%d) Can't create framebuffer object - unsupported format!",i);
    break;
   default:
    fatalf("(FBO%d) Can't create framebuffer object (ERROR=%d)!",i,status);
    return;
  }
 }
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
 glDrawBuffersARB(1,&screen_buf);
}

/** destroy FB objects */
void killFBO() {
 assert(use_fbo);
 printf("Deleting framebuffer objects\n");
 glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
 glDrawBuffersARB(1,&screen_buf);
 glDeleteFramebuffersEXT(2,fbuf);
 glDeleteRenderbuffersEXT(2,rbuf_depth);
 glDeleteTextures(2,fb_image);
 glerr("killFBO");
}

/** callback called when resizing window */
void resizeCallback(int x,int y) {
 maxx=x;
 maxy=y;
 if (use_fbo) {
  killFBO();
  initFBO(maxx,maxy);
 }
 glutPostRedisplay();
}

/** set 2d projection (0,0 - maxx,maxy), turn off Z-buffer and backface culling */
void set2D(void) {
 glerr("set2d begin");
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 glDisable(GL_DEPTH_TEST);
 glDisable(GL_CULL_FACE);
 glDisable(GL_LIGHTING);

 glScaled(2.0/(double)maxx,-2.0/(double)maxy,1.0);
 glTranslated(-((double)maxx/2.0),-((double)maxy/2.0),0.0);
 glViewport(0,0,maxx,maxy);
 material(NULL,NULL,NULL,white);//reset material
 glerr("set2d end");
}

int last_tick=gettickcount();

/** timer function called each frame */
void tick(void){
 int now=gettickcount();
 ms_last=now-last_tick;
 last_tick=now;
 radius+=(radiusdelta*ms_last)*kb_sensitivity/1000.0;
 radius=max(1,radius);
}

/** texture list.*/
typedef struct _texlist {
 texture *t;
 char *name;
 struct _texlist *next;
} tex_list;

/** texture list */
class texlist {
 tex_list *t_l;
public:
 texlist(){
  t_l=NULL;
 }
 /** get texture with given name. If not exist, load it */
 texture *gettex(const char *tname) {
  tex_list *t=t_l;
  while (t) {
   if (strcasecmp(t->name,tname)==0) return t->t;
   t=t->next;
  }
  //not found
  t=(tex_list *)malloc0(sizeof(tex_list));
  t->next=t_l;
  t->name=strdup(tname);
  if (strcasecmp(t->name,"$tex3d_1")==0) t->t=fs_3dtexture(t3[0]);
  else if (strcasecmp(t->name,"$tex3d_2")==0) t->t=fs_3dtexture(t3[1]);
  else t->t=fs_loadtexture(tname);
  printf("Loading: %s\n",tname);
  t_l=t;
  return t_l->t;
 }
};

/** One texture list for entire program */
texlist *tl=new texlist();

/** growable list of items */
class itemlist {
 item **items;
 int count;
 int capacity;
public:
 /** create new list with initial cpacity
  @param cap initial capacity (number of items) in list */
 itemlist(int cap=2) {
  capacity=cap;
  if (capacity<1) capacity=1;
  count=0;
  items=(item**)malloc(sizeof(item*)*capacity);
 }

 /** add item to list (at end) */
 void add(item *obj) {
  if (count==capacity) {
   capacity*=2;
   items=(item**)realloc(items,sizeof(item*)*capacity);
  }
  items[count]=obj;
  count++;
 }
 /** draw entire list (draw items in the order in which they are in list) */
 void draw() {
  for (int i=0;i<count;i++) {
   items[i]->draw();
  }
 }

 /** return: 1 if a is before b, -1 if a is after b, 0 if distance is equal */
 static int compare(const void* a,const void *b) {
  float da=(*(item**)a)->distance2(eye,eye_d);
  float db=(*(item**)b)->distance2(eye,eye_d);
//  printf("dist -> %f %f %x %x\n",da,db,a,b);
  if (da<db) return 1;
  if (da>db) return -1;
  return 0;
 }

 /** sort list by distance from eye (nearest last) */
 void sort() {
  int sorted=0,max=count-1;
  for (int i=0;i<count;i++) {
   items[i]->recalc_distance2();
  }
/*  printf("sorting %d\n",count);
  for (int i=0;i<count;i++) {
   printf("[%d] %x %g\n",i,items[i],items[i]->dist2);
  }*/
  //qsort unusable, it is not stable
  while(!sorted) {
   sorted=1;
   for (int i=0;i<max;i++) {
    if (items[i]->dist2<items[i+1]->dist2) {
     item *x=items[i+1];
     items[i+1]=items[i];
     items[i]=x;
     sorted=0;
    }
   }
   max--;
  }

//  qsort(items,count,sizeof(item*),compare);
/*  for (int i=0;i<count;i++) {
   printf("[%d] %x %g\n",i,items[i],items[i]->dist2);
  }*/
 }

};


void set3D();

/** use framebuffer numbered 0 or 1. Set fbo_tex texture to the content of second framebuffer
 -1: copy FBO 0 to screen, show screen */
void fbo_use(int fbo,int clear_fbo) {
 if (!use_fbo) return;
 if (fbo==-1) {
  fbo_tex=fb_image[0];
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
  glDrawBuffersARB(1,&screen_buf);
  set2D();
  glDisable(GL_BLEND);//needed, otherwise very strange effects occur
  glDisable(GL_DEPTH_TEST);
  glEnable(FBTYPE);
  glBindTexture(FBTYPE, fb_image[0]);
  glBegin(GL_TRIANGLE_STRIP);
  glTexCoord2f(0,1);
  glVertex3f(0,0,0);
  glTexCoord2f(0,0);
  glVertex3f(0,maxy,0);
  glTexCoord2f(1,1);
  glVertex3f(maxx,0,0);
  glTexCoord2f(1,0);
  glVertex3f(maxx,maxy,0);
  glEnd();
  glDisable(FBTYPE);  
 } else {
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbuf[fbo]);
  fbo_tex=fb_image[1-fbo];
  glDrawBuffersARB(1,fb_buf);
  if (clear_fbo) {
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  }
 } 
}

/** special "model" - switch framebuffer object to draw to
 */
class fbo_switch:public empty {
 int target,clear;
public:
 /** 
 target can be:
 0: first framebuffer (fb0), set fbo_tex to framebuffer 1 contents
 1: second framebuffer (fb1), set fbo_tex to framebuffer 0 contents
 -1: screen, copy framebuffer 0 to screen
 clear: clear framebuffer after switch?
 */
 fbo_switch(int _target,int _clear) : empty() {
  target=_target;
  clear=_clear;
 }
 /** switch framebuffer */
 void draw() {
  fbo_use(target,clear);
 }
};

char* ext;//seznam extensions

/** true - extension is supported. false - not supported
 @param what Extension to be queried for support */
int gl_ext(const char *what) {
 char *ex=ext;
 int nw=strlen(what);
 while (*ex) {
  if ((int)strcspn(ex," ")==nw && strncasecmp(what,ex,nw)==0) return 1;
  //next extension
  while (*ex!=' ' && *ex!=0) ex++;
  while (*ex==' ') ex++;
 }
 return 0;
}

#define check_ext(x) if (!x) fatalf(#x " required!");

/** initialize functions needed for EXT_framebuffer_object.
 Return false if no FBO support in card/drivers */
int init_fbo_func() {
  if (!gl_ext("GL_EXT_framebuffer_object"))
          fatalf("EXT_framebuffer_object not supported!");
  if (!gl_ext("GL_ARB_texture_non_power_of_two"))
          fatalf("ARB_texture_non_power_of_two not supported!");
  glGenFramebuffersEXT        =(PFNGLGENFRAMEBUFFERSEXTPROC)wglGetProcAddress("glGenFramebuffersEXT");
  glGenRenderbuffersEXT       =(PFNGLGENRENDERBUFFERSEXTPROC)wglGetProcAddress("glGenRenderbuffersEXT");
  glDeleteFramebuffersEXT     =(PFNGLDELETEFRAMEBUFFERSEXTPROC)wglGetProcAddress("glDeleteFramebuffersEXT");
  glDeleteRenderbuffersEXT    =(PFNGLDELETERENDERBUFFERSEXTPROC)wglGetProcAddress("glDeleteRenderbuffersEXT");
  glBindRenderbufferEXT       =(PFNGLBINDRENDERBUFFEREXTPROC)wglGetProcAddress("glBindRenderbufferEXT");
  glRenderbufferStorageEXT    =(PFNGLRENDERBUFFERSTORAGEEXTPROC    )wglGetProcAddress("glRenderbufferStorageEXT");
  glBindFramebufferEXT        =(PFNGLBINDFRAMEBUFFEREXTPROC        )wglGetProcAddress("glBindFramebufferEXT");
  glFramebufferTexture2DEXT   =(PFNGLFRAMEBUFFERTEXTURE2DEXTPROC   )wglGetProcAddress("glFramebufferTexture2DEXT");
  glDrawBuffersARB            =(PFNGLDRAWBUFFERSARBPROC            )wglGetProcAddress("glDrawBuffersARB");
  glFramebufferRenderbufferEXT=(PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)wglGetProcAddress("glFramebufferRenderbufferEXT");
  glCheckFramebufferStatusEXT =(PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC )wglGetProcAddress("glCheckFramebufferStatusEXT");
  check_ext(glGenFramebuffersEXT);
  check_ext(glGenRenderbuffersEXT);
  check_ext(glBindRenderbufferEXT);
  check_ext(glRenderbufferStorageEXT);
  check_ext(glBindFramebufferEXT);
  check_ext(glFramebufferTexture2DEXT);
  check_ext(glDrawBuffersARB);
  check_ext(glFramebufferRenderbufferEXT);
  check_ext(glCheckFramebufferStatusEXT);
  return 1;
 }



/** scene with multiple models */
class scene {
 /** ordinary items (draw in order in which they are added) */
 itemlist *oitems;
 /** transparent items (draw after all ordinary items, from back to front) */
 itemlist *titems;
 /** special items (draw after all ordinary and transparent items, from back to front) */
 itemlist *sitems;
 /** current line while loading scene */
 int line;
 /** current command while loading scene */
 char *cmd;
 /** current item in parser */
 item *current;
 /** print syntax error message and exit */
 void syntax_error(const char *message) {
  fatalf("Scene parse error line %d: Syntax '%s %s'\n",line,cmd,message);
 }
 /** exit if there is no curent object */
 void check_current() {
  if (!current) fatalf("Scene parse error line %d: No object for %s\n",line,cmd);
 }
 /** return parameter as int */
 int intparam(const char *tmp) {
  int va;
  int r=sscanf(tmp,"%d",&va);
  if (r!=1) fatalf("Scene parse error line %d: %s expecting integer parameter, found '%s'\n",line,cmd,tmp);
  return va;
 }

public:
 ~scene() {
  delete titems;
  delete oitems;
  delete sitems;
 }
 /** create new scene with initial cpacity
  @param cap initial capacity (number of ordinary items and transparent items) in each list */
 scene(int cap=2) {
  oitems=new itemlist(cap);
  titems=new itemlist(cap);
  sitems=new itemlist(cap);
 }
 /** add item to scene */
 void add(item *obj) {
  if (obj->transparent==1) {
   titems->add(obj);
  } else if (obj->transparent>1) {
   sitems->add(obj);
  } else {
   oitems->add(obj);
  }
 }
 /** draw entire scene (draw items in the order in which they were added) */
 void draw() {
  oitems->draw();
  titems->sort();
  titems->draw();
  sitems->sort();
  sitems->draw();
 }

 /** load the scene from definition file
  @returns true if scene successfully loaded
  @param filename name of the file to load
  */
 int load(const char *filename) {
  int needfbo=0;
  float v_move[3]={0};
  float v_dist[3]={0};
  float v_rotate[4]={0};
  shader sh=NULL,last_f=NULL,last_v=NULL;
  shader common_f=new tshader('f',"","");
  shader common_v=new tshader('v',"","");
  char shadertype=0;
  int l;
  char buf[2048];
  char param1[257],param2[257];
  float param[4];
  char *tmp;
  FILE *f=fopen(filename,"r");
  if (!f) return 0;
  line=0;current=NULL;
  while (fgets(buf,sizeof(buf),f)) {
   line++;
   cmd=buf;
   l=strlen(cmd);
   for (int j=0;j<l;j++) if (cmd[j]=='\t') cmd[j]=' ';//tabs to spaces
   while (l) { //remove trailing whitespace;
    if (cmd[l-1]==10 || cmd[l-1]==' ' || cmd[l-1]==13) {
     cmd[l-1]=0;l--;
    } else break;
   }
   while (cmd[0]==' ') cmd++;
   if (cmd[0]=='#') continue;
   if (!strlen(cmd)) continue;
   tmp=strchr(cmd,' ');
   if (tmp) {
    tmp[0]=0;
    tmp++;
    while (tmp[0]==' ') tmp++;
   }  
   //printf("CMD %s %s\n",cmd,tmp);
   if (strcasecmp(cmd,"empty")==0) {	//COMMAND
    current=new empty();
   } else if (strcasecmp(cmd,"pseudocylinder")==0) {	//COMMAND
    int xseg,yseg;
    float w,h,exp;
    int r=sscanf(tmp,"%d %d %f %f %f",&xseg,&yseg,&w,&h,&exp);
    if (r!=5) syntax_error("Xseg Yseg Width Height Exp");
    current=new pseudocylinder(xseg,yseg,w,h,exp);
   } else if (strcasecmp(cmd,"cone")==0) {	//COMMAND
    int xseg,yseg;
    float w0,w1,h;
    int r=sscanf(tmp,"%d %d %f %f %f",&xseg,&yseg,&w0,&w1,&h);
    if (r!=5) syntax_error("Xseg Yseg WidthUp WidthDown Height");
    current=new coneseg(xseg,yseg,w0,w1,h);
   } else if (strcasecmp(cmd,"flat")==0) {	//COMMAND
    float x1,y1,x2,y2,z;
    int r=sscanf(tmp,"%f %f %f %f %f",&x1,&y1,&x2,&y2,&z);
    if (r!=5) syntax_error("X1 Y1 X2 Y2 Z");  
    current=new flat(x1,y1,x2,y2,z);
   } else if (strcasecmp(cmd,"cylinder")==0) {	//COMMAND
    int xseg,yseg;
    float w,h;
    int r=sscanf(tmp,"%d %d %f %f",&xseg,&yseg,&w,&h);
    if (r!=4) syntax_error("Xseg Yseg Width Height");
    current=new cylinder(xseg,yseg,w,h);
   } else if (strcasecmp(cmd,"cylindercaps")==0) {	//COMMAND
    int xseg;
    float w,h;
    int r=sscanf(tmp,"%d %f %f",&xseg,&w,&h);
    if (r!=3) syntax_error("Xseg Width Height");
    current=new cylindercaps(xseg,w,h);
   } else if (strcasecmp(cmd,"house")==0) {	//COMMAND
    current=new house();
   } else if (strcasecmp(cmd,"fbo")==0) {	//COMMAND
    int fbonum,clearbit;
    int r=sscanf(tmp,"%d %d",&fbonum,&clearbit);
    if (r!=2 ||fbonum<-1 || fbonum>1) syntax_error("-1|0|1 0|1");
    current=new fbo_switch(fbonum,clearbit);
   } else if (strcasecmp(cmd,"color")==0) {	//COMMAND
    int col[4]={0};
    ub4 cl;
    int r=sscanf(tmp,"%d %d %d %d",&col[0],&col[1],&col[2],&col[3]);
    if (r==3) col[3]=255;
    if (r>4 || r<3) syntax_error("( R G B | R G B A) (values between 0 and 255)");
    for (int i=0;i<4;i++) cl[i]=col[i];
    current->setcolor(cl);   
   } else if (strcasecmp(cmd,"texture")==0) {	//COMMAND
    check_current();
    current->settexture(tl->gettex(tmp));
   } else if (strcasecmp(cmd,"move")==0) {	//COMMAND
    check_current();
    if (strcasecmp(tmp,"last")!=0) {
     int r=sscanf(tmp,"%f %f %f",&v_move[0],&v_move[1],&v_move[2]);
     if (r!=3) syntax_error("(X Y Z | last)");
    }
    current->move(v_move[0],v_move[1],v_move[2]);
   } else if (strcasecmp(cmd,"distance")==0) {	//COMMAND
    check_current();
    if (strcasecmp(tmp,"last")!=0) {
     int r=sscanf(tmp,"%f %f %f",&v_dist[0],&v_dist[1],&v_dist[2]);
     if (r!=3) syntax_error("(X Y Z | last)");
    }
    current->setdistance(v_dist[0],v_dist[1],v_dist[2]);
   } else if (strcasecmp(cmd,"rotate")==0) {	//COMMAND
    check_current();
    if (strcasecmp(tmp,"last")!=0) {
     int r=sscanf(tmp,"%f %f %f %f",&v_rotate[0],&v_rotate[1],&v_rotate[2],&v_rotate[3]);
     if (r!=4) syntax_error("(Angle X Y Z | last)");
    }
    current->rotate(v_rotate[0],v_rotate[1],v_rotate[2],v_rotate[3]);
   } else if (strcasecmp(cmd,"camdistance")==0) {	//COMMAND
    radius=atof(tmp);
   } else if (strcasecmp(cmd,"nearclip")==0) {		//COMMAND
    clipfront=atof(tmp);
   } else if (strcasecmp(cmd,"farclip")==0) {		//COMMAND
    clipback=atof(tmp);
   } else if (strcasecmp(cmd,"sameshader")==0) {	//COMMAND
    int r=sscanf(tmp,"%c",&shadertype);
    if (r>=1 && (shadertype=='f' || shadertype=='v')) {
     check_current();
     if (shadertype=='f') sh=last_f; else sh=last_v;
     if (!sh) fatalf("Scene parse error line %d: No shader to add\n",line);
     current->setshader(shadertype,sh);
     sh=NULL;
    } else syntax_error("(f|v) (you must specify vertex/fragment shader in first parameter)");
   } else if (strcasecmp(cmd,"shader")==0) {		//COMMAND
    int r=sscanf(tmp,"%c %256s %256s",&shadertype,param1,param2);
    if (r>=3 && (shadertype=='f' || shadertype=='v')) {
     sh=new tshader(shadertype,param1,param2);
     //store as last shader
     if (shadertype=='f') {
      last_f=sh;
      common_f->putparams(sh);
     } else {
      last_v=sh;
      common_v->putparams(sh);
     }
    } else syntax_error("(f|v) File Function (you must specify vertex/fragment shader in first parameter)");
   } else if (strcasecmp(cmd,"param")==0) {		//COMMAND
    if (!sh) fatalf("Scene parse error line %d: No shader for parameter\n",line);
    int r;
    r=sscanf(tmp,"%256s $%256s",param1,param2);
    if (r==2) {
     sh->addparam(param1,param2);
    } else {
     r=sscanf(tmp,"%256s %f %f %f %f",param1,&param[0],&param[1],&param[2],&param[3]);
     if (r>=2) {
      sh->addparam(param1,r-1,param);
     } else syntax_error("NAME ($VARIABLE | FLOAT (FLOAT) (FLOAT) (FLOAT))");
    }
   } else if (strcasecmp(cmd,"commonparam")==0) {		//COMMAND
    int r;
    r=sscanf(tmp,"%c ",&shadertype);//first parameter
    if (r>=1 && (shadertype=='f' || shadertype=='v' || shadertype=='*')) {
    } else syntax_error("(f|v|*) NAME ($VARIABLE | FLOAT (FLOAT) (FLOAT) (FLOAT)) (you must specify vertex/fragment shader in first parameter)");
    r=sscanf(tmp,"%c %256s $%256s",&shadertype,param1,param2);
    if (r==3) {
     if (shadertype!='f') common_v->addparam(param1,param2);
     if (shadertype!='v') common_f->addparam(param1,param2);
    } else {
     r=sscanf(tmp,"%c %256s %f %f %f %f",&shadertype,param1,&param[0],&param[1],&param[2],&param[3]);
     if (r>=3) {
      if (shadertype!='f') common_v->addparam(param1,r-1,param);
      if (shadertype!='v') common_f->addparam(param1,r-1,param);
     } else syntax_error("NAME ($VARIABLE | FLOAT (FLOAT) (FLOAT) (FLOAT))");
    }
   } else if (strcasecmp(cmd,"addshader")==0) {	//COMMAND
    check_current();
    if (!sh) fatalf("Scene parse error line %d: No shader to add\n",line);
    //add common parameters
    shadertype=sh->type();
    current->setshader(shadertype,sh);
    sh=NULL;
   } else if (strcasecmp(cmd,"cull")==0) {	//COMMAND
    check_current();
    current->cull=intparam(tmp);
   } else if (strcasecmp(cmd,"transparent")==0) {	//COMMAND
    check_current();
    current->transparent=intparam(tmp);
   } else if (strcasecmp(cmd,"zwrite")==0) {	//COMMAND
    check_current();
    current->zwrite=intparam(tmp);
   } else if (strcasecmp(cmd,"needfbo")==0) {	//COMMAND
    check_current();
    needfbo=1;
   } else if (strcasecmp(cmd,"depthtest")==0) {	//COMMAND
    check_current();
    current->d_test=intparam(tmp);
   } else if (strcasecmp(cmd,"addobj")==0) {	//COMMAND
    check_current();
    if (needfbo && !use_fbo) { //discard
     delete current;
    } else {
     add(current);
    }
    current=NULL;
    needfbo=0;
   } else {
    fatalf("Scene parse error line %d: Unknown command '%s'\n",line,cmd);
   }
  }
  fclose(f);
  delete common_f;
  delete common_v;
  return 1;
 }
};

scene *world;

/** load entire wirld, generate textures, etc ... */
void loadWorld() {
 //generate 3d textures
 printf("Loading world...\n");
 glGenTextures(2,t3);
 for (int i=0;i<2;i++) {
  generate3d(&t3data[i]);
  glPixelStorei(GL_UNPACK_ALIGNMENT,8);
  glBindTexture(GL_TEXTURE_3D, t3[i]);
  glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glerr("loadWorld - glTexImage3D0");
  glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glerr("loadWorld - glTexImage3D1");
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glerr("loadWorld - glTexImage3D2");
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glerr("loadWorld - glTexImage3D3");
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_REPEAT);
  glerr("loadWorld - glTexImage3D4");
  glTexImage3D(GL_TEXTURE_3D, 0,GL_RGBA, DIM3D, DIM3D, DIM3D, 0,GL_RGBA,GL_UNSIGNED_BYTE,t3data[i]);
  glerr("loadWorld - glTexImage3D");
 }
 world=new scene();
 if (!world->load(scene_filename)) fatalf("Cannot load scene file: '%s'",scene_filename);
 printf("Finished loading world...\n");
}

int last_ftick=gettickcount();
int last_frames=0;

/** conpute fps information at end of frame */
void fpsFrameEnd() {
 int now=gettickcount();
 last_frames++;
 if (now>last_ftick+500) {
  ms_curr=(now-last_ftick)/(float)last_frames;
  last_frames=0;
  last_ftick=now;
 }
}

/** cross product of 2 vectors (3 component vectors)
 @param r result of dot product (out)
 @param v1 first vector (in)
 @param v2 second vector (in)
 */
void cross (float *r,float *v1,float *v2) {
 r[0]=v1[1]*v2[2]-v2[1]*v1[2];
 r[1]=v1[2]*v2[0]-v2[2]*v1[0];
 r[2]=v1[0]*v2[1]-v2[0]*v1[1];
}

/** draw HUD (including fps information)*/
void drawHud(void) {
 set2D();
 if (help) {
  gltext_at("Flame version " VERSION "\nh : toggle this help\nq : quit program\nw : toggle wireframe display\nmouse : rotate view\nUP : move forward\nDOWN : move backward",5,5);
 } else {
  gltext_at("h=help,q=quit",5,5);
 }
 char fps[256];
 sprintf(fps,"%2.6f ms, %6.2f fps %dx%d",(float)ms_curr,1000.0/ms_curr,maxx,maxy);
 gltext_at(fps,5,maxy-5-fontsizey);
}

/** set 3d projection */
void set3D() {
 float3 eye_left;
 glerr("set3d begin");
 glEnable(GL_DEPTH_TEST);
 glEnable(GL_CULL_FACE);
 glMatrixMode(GL_PROJECTION);
 glEnable(GL_COLOR_MATERIAL);
 glLoadIdentity();
 gluPerspective(fov,(float)maxx/(float)maxy,clipfront,clipback);
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();
 eye[0]=sin(angle1)*cos(angle2)*radius;
 eye[1]=cos(angle1)*cos(angle2)*radius;
 eye[2]=sin(angle2)*radius;
 eye_d[0]=-eye[0];
 eye_d[1]=-eye[1];
 eye_d[2]=-eye[2];
 eye_u[0]=0;
 eye_u[1]=0;
 eye_u[2]=1;
 cross(eye_left,eye_u,eye_d);
 cross(eye_u,eye_d,eye_left);
 normalize3(eye_d);
 normalize3(eye_u);
// printf("%f %f %f %f %f %f\n",eye_d[0],eye_d[1],eye_d[2],eye_u[0],eye_u[1],eye_u[2]);
 gluLookAt(eye[0],eye[1],eye[2]
          ,0,0,0
          ,eye_u[0],eye_u[1],eye_u[2]);//nastaveni pohledu
 glViewport(0,0,maxx,maxy);
 glerr("set3d end");
}

/** set up lighting */
void setLight() {
 GLfloat pos[4]={50,50,50,1};
 GLfloat dif[4]={0.5,0.5,0.5,1};
 GLfloat amb[4]={0.1,0.1,0.1,1};
 glerr("light begin");
 glLightfv(GL_LIGHT0,GL_POSITION,pos);
 glLightfv(GL_LIGHT0,GL_AMBIENT, amb);
 glLightfv(GL_LIGHT0,GL_DIFFUSE, dif);
 glLightfv(GL_LIGHT0,GL_SPECULAR,dif);
 glEnable(GL_LIGHT0);
 glEnable(GL_LIGHTING);
 glerr("light end");
}

/** draw entire world (scene, lighting ... ) */
void drawWorld() {
 glerr("drawworld begin");
 set3D();
 setLight();
 glerr("set3d");
 world->draw();
 glerr("drawworld end");
}

/** callback used when needed to update display */
void displayCallback(void) {
 glClearColor(fogcolor[0],fogcolor[1],fogcolor[2],0.0);
 glEnable(GL_DEPTH_TEST);
 glDepthRange(0,1);
 if (!use_fbo) glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 glDepthFunc(GL_LESS);
 glClear(GL_COLOR_BUFFER_BIT);
 drawWorld();
 drawHud();
 glFlush();
 glutSwapBuffers();
 fpsFrameEnd();
}

/** idle callback (called when application is idle) */
void idleHandler(void){
 tick();
 glutPostRedisplay();
}

/** register all callbacks */
void setCallbacks (void) {
 glerr("callback begin");
 //grafika
 glutDisplayFunc(displayCallback);
 glutReshapeFunc(resizeCallback);
 glutIdleFunc(idleHandler);
// glutWindowStatusFunc(b_win);
 glutMotionFunc(mouseHandler);
 glutPassiveMotionFunc(mouseHandler);
 glutMouseFunc(mouseButtonHandler);
 glutSetCursor(GLUT_CURSOR_NONE);
 mouseCenter();
 //klavesnice
 glutKeyboardFunc(keyDownHandler);
 glutKeyboardUpFunc(keyUpHandler);
 glutSpecialFunc(skeyDownHandler);
 glutSpecialUpFunc(skeyUpHandler);
 glutIgnoreKeyRepeat(1);
 glerr("callback end");
}

/** set up small details based on platform */
void windowInit(void){
 #ifdef win32
 HANDLE h,m;
 //aby okno neslo zavrit -> kdyz jde, tak to sleti
 h=FindWindow("GLUT",apptitle);
 (void)SetCapture(h);
 (void)SetFocus(h);
 m=GetSystemMenu(h,false);
 (void)EnableMenuItem(m,SC_CLOSE,MF_BYCOMMAND | MF_DISABLED | MF_GRAYED);
 #endif
}

/** width of window */
int width=800;
/** height of window */
int height=600;
/** attempt fullscreen? */
int fullscreen=0;

/** initialize OpenGL, create window ...
 @param w window width
 @param h window height
 @param argc argc passed to GLUT
 @param argv argv passed to GLUT
 */
void createWindow(int w,int h,int *argc, char** argv) {
 glutInit(argc,argv);
 glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
 glutInitWindowSize(w,h);
 glutInitWindowPosition(0,0);
 gwin=glutCreateWindow(apptitle);
 if (fullscreen) glutFullScreen();
 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
 ext=(char*)glGetString(GL_EXTENSIONS);
 if (gl_ext("GL_NV_fog_distance")) {
  glFogi(FOG_DISTANCE_MODE_NV,EYE_RADIAL_NV); //lepe vypadajici mlha
 }
 if (use_fbo) if (!init_fbo_func()) use_fbo=0;
 glTexImage3D=(PFNGLTEXIMAGE3DPROC)wglGetProcAddress("glTexImage3D");
 if (!glTexImage3D) fatalf("glTexImage3D required!");
 maxx=w;
 maxy=h;
}

/** structure containing definition of one configuration element */
typedef struct {
 char* key; //key
 void *value; //pointer to value
 int typ; //type of value (i=int, f=float, s=string)
 float min; //minimal acceptable value
 float max; //maximal acceptable value
} config;

/** configuration data definition */
config cfg[]={
{"width",&width,'i',320,4096},
{"height",&height,'i',200,4096},
{"fullscreen",&fullscreen,'i',0,1},
{"font_x",&fontsizex,'f',8,64},
{"font_y",&fontsizey,'f',16,64},
{"sensitivity",&sensitivity,'f',0.001,1000},
{"kbsensitivity",&kb_sensitivity,'f',0.001,1000},
{"fov",&fov,'f',5,175},
{"font",&font_texture_name,'s',0,0},
{"scene",&scene_filename,'s',0,0},
{"usefbo",&use_fbo,'i',0,1},
};

/** load configuration from a file*/
void loadConfig(const char *filename) {
 char buf[1024];
 char *s,*tmp;
 FILE *f=fopen(filename,"r");
 if (!f) return;	//Can't load config - use defaults
 while (fgets(buf,sizeof(buf),f)) {
  s=buf;
  while (s[0]==' ' || s[0]=='\t') s++;
  if (s[0]=='#') continue;
  if (!strlen(s)) continue;
  tmp=strchr(s,'=');
  if (!tmp) continue;
  tmp[0]=0;
  tmp++;
  for (uint i=0;i<elements(cfg);i++) {
   if (strcasecmp(s,cfg[i].key)==0) {
    if (cfg[i].typ=='i') {		//int
     int val=atoi(tmp);
     if (val<cfg[i].min) val=(int)cfg[i].min;
     if (val>cfg[i].max) val=(int)cfg[i].max;
     (*((int*)cfg[i].value))=val;
    } else if (cfg[i].typ=='s') {	//string
     int q=(int)strlen(tmp);
     while (q>0 && (tmp[q-1]==13 || tmp[q-1]==10)) {
      tmp[q-1]=0;q--;
     }
     free(*((char **)cfg[i].value));
     *((char **)cfg[i].value)=strdup(tmp);
    } else if (cfg[i].typ=='f') {	//float
     float val=atof(tmp);
     if (val<cfg[i].min) val=cfg[i].min;
     if (val>cfg[i].max) val=cfg[i].max;
     (*((float*)cfg[i].value))=val;
    } else {
     fatalf("loadconfig: internal error\n");
    }
   }
  }
 }
 fclose(f);
}

/** main function*/
int main(int argc,char** argv){
 srand(gettickcount());
 loadConfig("config.txt");
 createWindow(width,height,&argc,argv);
 setCallbacks();
 windowInit();
 printf("Loading font: %s\n",font_texture_name);
 tex_font=fs_loadtexture(font_texture_name);
 if (use_fbo) initFBO(maxx,maxy);
 createCg();
 loadWorld();
 glutShowWindow();
 glutMainLoop();//this will never return
 return 0;
}

#ifdef WIN32

/** main funkce pro GDI aplikace */
int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow) {
 return main (1,&lpCmdLine);//neni to presne, ale tim co mi pristane na prikazove radce se stejne nezabyvam (zatim)
};

#endif

