#include <sys/stat.h>

// Na windows a linuxu jsou jine oddelovace cesty
#ifdef WIN32
#define fs '\\'		//nativni filesystem separator
#define fs2 '/'		//alternativni filesystem separator
#else
#define fs '/'		//nativni filesystem separator
#define fs2 '\\'	//alternativni filesystem separator
#endif

/** debugging routine to write opengl errors */
#ifndef NDEBUG
void glerr(char *x) {
 int e=glGetError();//cislo chyby
 char *ee=0;//popis chyby
 if (!e) return;
 if (e==GL_INVALID_ENUM) ee="GL_INVALID_ENUM";
 if (e==GL_INVALID_VALUE) ee="GL_INVALID_VALUE";
 if (e==GL_INVALID_OPERATION) ee="GL_INVALID_OPERATION";
#ifdef GL_TABLE_TOO_LARGE
 if (e==GL_TABLE_TOO_LARGE) ee="GL_TABLE_TOO_LARGE";
#endif
 if (e==GL_OUT_OF_MEMORY) ee="GL_OUT_OF_MEMORY";
 if (e==GL_STACK_OVERFLOW) ee="GL_STACK_OVERFLOW";
 if (e==GL_STACK_UNDERFLOW) ee="GL_STACK_UNDERFLOW";
 if (e==GL_INVALID_ENUM) ee="GL_INVALID_ENUM";
 if (ee) printf("GL_ERROR at %s :%d %s\n",x,e,ee);
  else printf("GL_ERROR at %s :%d\n",x,e);
}
#else
#define glerr(dummy) //do nothing
#endif

typedef struct {//hlavicka TGA souboru.
 char idbytes;     //0 je spravne, jinak-problem (na miste pixelu jsou navic jina odpadni data -> problem uzivatele)
 char colormap;    //lze ignorovat
 char type;	    //pouze 2 je spravne (RGB)
 char colmap[5];   //lze ignorovat
 short orgx,orgy;  //pocatek obrazku, ignouje se
 short x,y;        //rozmery obrazku
 char bpp;         //bits per pixel
 char flags;       //nezajimava polozka
} tgaheader;

typedef struct { //TGA soubor. Hlavicka + data
 tgaheader head;	//hlavicka tga souboru
 unsigned char data[1];	    //x*y*(bpp/8) bajtu
} tgafile;

/** jako strncpy, ale posledni znak vzdy nastavi na NULL */
char *b_strncpy(char *dest,const char *source, size_t count) {
 char *q=strncpy(dest,source,count);
 q[count-1]=0;
 return q;
}

/** jako strncat, ale pocet znaku je omezeni delky retezce celkem, vcetne NULL znaku, ne jen pridavaneho retezce */
char *b_strncat(char *dest,const char *source, size_t count) {
 return strncat(dest,source,count-strlen(dest)-1);
}

/** vrati velikost souboru */
int fsize(const char *f) {
 struct stat b;
 if (stat(f,&b)) return 0;//return 0 if error
 return b.st_size;
}

/** otevre soubor a vrati jeho velikost */
FILE* fopen2(const char *s1,const char *s2,const char *m, uint *filesize){
 FILE *f;
 char str[257];
 char *t;
 b_strncpy(str,s1,sizeof(str));
 b_strncat(str,s2,sizeof(str));
 t=str;
 while (*t) {
  if (*t==fs2) *t=fs;
  t++;
 }
 f=fopen(str,m);
 if (!filesize) return f;
 if (f) *filesize=fsize(str); else *filesize=0;
 return f;
}

/** nahraje soubor do pameti, vrati velikost */
int fs_loadfile(void **out,const char *name) {
 FILE *f;
 uint s=0;
 f=fopen2("data/",name,"rb",&s);
 if (!f) {
  *out=0;
  return 0;
 }
 if (s) {
  *out=malloc(s);
  if (*out) {
   s=fread(*out,1,s,f);
  }
 } else (*out)=(void *)0xdeadbeef;//"naalokovany pointer o velikosti 0" -> stejne na nej nejde hrabnout, staci kdyz neni null
 (void)fclose(f);
 return s;
}

/** nahraje texturu z pameti do pameti graficke karty, k texture navic vyrobi mipmapy */
void gl_loadtex(int num,int bpp,int x,int y,void *tex) {
 glPixelStorei(GL_UNPACK_ALIGNMENT,8);
 glBindTexture(GL_TEXTURE_2D, num);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
 if (gluBuild2DMipmaps(GL_TEXTURE_2D,bpp,x,y,(bpp==3)?GL_RGB:GL_RGBA,GL_UNSIGNED_BYTE,tex))
 fatalf("gl_loadtex: Unable to build mipmaps for texture %d (x=%d, y=%d, bpp=%d)",num,x,y,bpp); 
}

/** nahraje texturu v pameti v .tga formatu do pameti graficke karty. Detaily viz loadtex */
void gl_loadtgatex(int num,tgafile *tex) {
 if (!tex) fatalf("gl_loadtgatex: Texture %d not loaded!",num);
 gl_loadtex(num,(tex->head.bpp)/8,tex->head.x,tex->head.y,&(tex->data[0]));
}

/** Targy jsou v BGR misto RGB ... tohle je zkonvertuje do spravneho formatu*/
void fs_rgbfix(tgafile *t) {
 int i=0;
 unsigned char q;
 int size=(t->head.x)*(t->head.y)*(t->head.bpp/8);
 for (i=0;i<size;i+=t->head.bpp/8) {
  q=t->data[i];
  t->data[i]=t->data[i+2];
  t->data[i+2]=q;
 }
}

/** Spravi transparenci u textur. Pokud je vsude 0, tak ji nastavi vsude na 255*/
void fs_alphafix(tgafile *t) {
 int i,size;
 uint *p;
 if (t->head.bpp!=32) return;
 size=(t->head.x)*(t->head.y);
 p=(uint *)t->data;
 for (i=0;i<size;i++) {
  if (p[i] & 0xff000000) return;
 }
 for (i=0;i<size;i++) {
  p[i]|=0xff000000;
 }
}


/** nahraje tga texturu, vratin na ni pointer nebo NULL pri neuspechu */
tgafile *fs_loadtgafile(const char *name) {
 void *ou;
 int q=fs_loadfile(&ou,name);
 if (q<(int)sizeof(tgaheader)) { //neni to TGA, takze nic ...
  if(q && ou) free(ou);
  return NULL;
 }
 if (q<(int)(((tgaheader*)ou)->x*((tgaheader*)ou)->y*(((tgaheader*)ou)->bpp/8)+sizeof(tgaheader))) {
  //je to zakomprimovany, ci co ... zamitnout!
  fatalf("zakompreseny textury neumim (%s)!",name);
  if(q && ou) free(ou);
  return NULL;
 }
 if (ou) fs_rgbfix((tgafile*)ou);
 if (ou) fs_alphafix((tgafile*)ou);
 return (tgafile*)ou;
}

typedef struct { // textura
 uint id; //OpenGL handle textury
 int is3d; //jestli je nebo neni 3d texturou
 tgafile *mem; //kde je v pameti
} texture;

texture *failsafetex;//defaultni textura
texture *tex_font;

/** vyrobi "defaultni" hnusnou texturu, ktera je pouzita kdyz textura chybi/nelze nahrat
 Je to tam aby to kvuli tomu nesletelo, ale textura je tak hnusna, ze toho ze tam chybi
 si asi kazdy hned vsimne */
void fs_failsafe_texture(texture *tex) {
 int i,j;
 tgafile *t;
 //debug_printf("making failsafe texture\n");
 if (tex->mem) return;//textura jiz existuje - nic se nedeje
 t=(tgafile *)malloc(16*16*3+sizeof(tgaheader));
 tex->mem=t;
 tex->id=0;
 tex->is3d=0;
 t->head.x=16;
 t->head.y=16;
 t->head.bpp=3*8;
 for(i=0;i<16;i++) for(j=0;j<16;j++) {
  t->data[(i*16+j)*3]=((i==0)||(i==15))*255;
  t->data[(i*16+j)*3+1]=((j==0)||(j==15))*255;
  t->data[(i*16+j)*3+2]=((i+j)&1)*255;
 }
 //debug_printf("made failsafe texture\n");
}

/** naalokuje pamet a vyplni nulama.  Vrati 0 pri selhani */
void *malloc0(size_t t){
 void *c;
 c=malloc(t);
 if (c) memset(c,0,t);
 return c;
}

/** Nahraje texturu. Zvlada pouze TGA. Vrati jeji ID */
texture *fs_loadtexture(const char *name) {
 texture *ou=(texture*)malloc(sizeof(texture));
 if (!failsafetex) failsafetex=(texture*)malloc0(sizeof(texture));
 if (!failsafetex->mem) {// vyrobi defaultni hnusnou texturu
  fs_failsafe_texture(failsafetex);
 }
 ou->mem=fs_loadtgafile(name);//try tga
 ou->is3d=0;
 if (!ou->mem) {//return default ugly texture
  printf("%s cannot be loaded, used failsafe\n",name);
  ou->mem=failsafetex->mem;
 }
 glGenTextures(1,&ou->id);
 gl_loadtgatex(ou->id,ou->mem);
 return ou;
}

/** Vytvori 3d texturu */
texture *fs_3dtexture(GLuint idtex) {
 texture *ou=(texture*)malloc(sizeof(texture));
 ou->mem=0;//no direct data
 ou->is3d=1;
 ou->id=idtex;
 return ou;
}

float fontsizex=12;    // velikost fontu
float fontsizey=16;    // velikost fontu

#define xdelta (1.0/1024.0)
#define ydelta (-1.0/256.0)
/** vyplivne na obrazovku jeden znak a posune matici */
void glchar(int c) {//32x8 znaku ...
 float y=floor(c/32)/8.0;
 float x=((float)(c & 0x1f))/32.0;
 if (c==13) return;
 if (c==10) return;
 glerr("glchar:begin");
 glEnable(GL_BLEND);
 glBegin(GL_QUADS);
 glTexCoord2f(x+xdelta,1.0-y-1.0/8.0-ydelta);	  glVertex3f(0.0,fontsizey,0.0);
 glTexCoord2f(x+xdelta,1.0-y+ydelta);		  glVertex3f(0.0,0.0,0.0);
 glTexCoord2f(x-xdelta+1.0/32.0,1-y+ydelta);      glVertex3f(fontsizex,0.0,0.0);
 glTexCoord2f(x-xdelta+1.0/32.0,1-y-1.0/8.0-ydelta);glVertex3f(fontsizex,fontsizey,0.0);
 glEnd();
 glTranslatef(fontsizex,0.0,0.0);
 glerr("glchar:end");
}

/** vyplivne na obrazovky retezec, rozeznama znak newline*/
void gltext(const char *s) {
 size_t t;
 if (strlen(s)==0) return;
 glerr("text:begin");
 glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D,tex_font->id);
 glPushMatrix();
 glPushMatrix();
 for (t=0;t<strlen(s);t++) {
  if (s[t]=='\n') {
   glPopMatrix();
   glTranslatef(0.0,fontsizey,0.0);
   glPushMatrix();
  }
  glchar(s[t]);
 }
 glPopMatrix();
 glPopMatrix();
 glerr("text:end");
}

/** vyplivne text na obrazovce posunuty o zadane souradnice */
void gltext_at(const char *s,float x,float y) {
 glPushMatrix();
 glTranslatef(x,y,0);
 gltext(s);
 glPopMatrix();
}
